#include <assert.h>

#include <pro.h>
#include <ua.hpp>

#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPStaticAnalyzer.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"
#include "interfaces/idapro/all.h"

using namespace std;

STARS_InstructionID_t STARS_IDA_Instruction_t::GetNextInstructionID(void)
{
	STARS_ea_t addr = (STARS_ea_t) m_id; // invoke constructor to convert to address

	insn_t SMPcmd ;
	uint32 SMPfeatures;
	SMPGetCmd(addr, SMPcmd, SMPfeatures);
	STARS_ea_t next_addr = addr + SMPcmd.size;

	return STARS_InstructionID_t(next_addr);
};

STARS_InstructionID_t STARS_IDA_Instruction_t::GetTargetInstructionID(void) {
	assert(NULL != this->STARScmd.Operands[0]);
	STARS_ea_t TargetAddr = this->STARScmd.Operands[0]->GetAddr();
	return STARS_InstructionID_t(TargetAddr);
}

// Get instruction info by address from IDA Pro.
bool STARS_IDA_Instruction_t::STARS_GetCmd(void) {
	bool success = true;
	int InstrLen = 0;

#if 0
	// Fill cmd structure with disassembly of instr
#if IDA_SDK_VERSION < 600
	InstrLen = ua_ana0((STARS_ea_t) m_id);
#else
	InstrLen = decode_insn((STARS_ea_t) m_id);
#endif
#endif

	// Copy cmd fields to member STARScmd.
	this->STARScmd.itype = cmd.itype;
	this->STARScmd.size = cmd.size;
	this->STARScmd.auxpref = cmd.auxpref;
	this->STARScmd.segpref = cmd.segpref;
	this->STARScmd.insnpref = cmd.insnpref;
	this->STARScmd.flags = cmd.flags;

	if (0 >= InstrLen) {
		SMP_msg("ERROR: decode_insn failed at %lx \n", (unsigned long) m_id);
		this->STARScmd.size = 0;
		success = false;
		return success;
	}

	// Get the canonical features into member STARSfeatures.
	this->STARSfeatures = cmd.get_canon_feature();

	for (int i = 0; i < UA_MAXOP; ++i) {
		this->STARScmd.Operands.push_back(std::make_shared<STARS_IDA_op_t>(cmd.Operands[i]));
		this->STARScmd.Operands[i]->SetSpecFlag4(0); 
#ifdef __EA64__
		if (STARS_ISA_Bitwidth == 64) {
			// Copy the cmd.rex prefix into the op_t.specflag4 field for each operand
			//  that has a SIB byte.
			this->STARScmd.Operands[i]->SetSpecFlag4(this->STARScmd.rex);
		}
#endif
		// See comments on STARS_VEXPR and STARS_VSIB in SMPDataFlowAnalysis.h.
		//  These bits do not (as of IDA Pro 6.4) conflict with cmd.rex bits.
		if ((cmd.auxpref & aux_vexpr) != 0) {
			this->STARScmd.Operands[i]->SetBitInSpecFlag4(STARS_VEXPR);
		}

		switch (this->STARScmd.itype) {
		    case NN_vgatherdps:
			case NN_vgatherdpd:
			case NN_vgatherqps:
			case NN_vgatherqpd:
			case NN_vpgatherdd:
			case NN_vpgatherdq:
			case NN_vpgatherqd:
			case NN_vpgatherqq:
			  this->STARScmd.Operands[i]->SetBitInSpecFlag4(STARS_VSIB);
			default:
				;
		}
	}

	return success;
} // end of STARS_IDA_Instruction_t::STARS_GetCmd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeVoidOpnd(void) const {
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(InitOp));
} // end of STARS_IDA_Instruction_t::MakeVoidOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeImmediateOpnd(uintptr_t value) const {
	op_t TempOp = InitOp;
	TempOp.type = o_imm;
	TempOp.value = (uval_t) value;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeVoidOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeRegOpnd(uint16 RegNum) const {
	op_t TempOp = InitOp;
	TempOp.type = o_reg;
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeFloatingPointRegOpnd(uint16 RegNum) const {
	op_t TempOp = InitOp;
	TempOp.type = o_fpreg;
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeFloatingPointRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMMXRegOpnd(uint16 RegNum) const {
	op_t TempOp = InitOp;
	TempOp.type = o_mmxreg;
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMMXRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeXMMRegOpnd(uint16 RegNum) const {
	op_t TempOp = InitOp;
	TempOp.type = o_xmmreg;
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeXMMRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeYMMRegOpnd(uint16 RegNum) const {
	op_t TempOp = InitOp;
	TempOp.type = o_ymmreg;
	TempOp.reg = RegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeYMMRegOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMemPhraseOpnd(uint16 BaseRegNum, uint16 IndexRegNum, uint16_t ScaleFactor) const {
	// TODO: Construct SIB byte when IndexRegNum is used.
	op_t TempOp = InitOp;
	TempOp.type = o_phrase;
	TempOp.reg = BaseRegNum;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMemPhraseOpnd()

STARSOpndTypePtr STARS_IDA_Instruction_t::MakeMemDisplacementOpnd(uint16 BaseRegNum, uint16 IndexRegNum, uint16_t ScaleFactor, uintptr_t offset) const {
	// TODO: Construct SIB byte when IndexRegNum is used.
	op_t TempOp = InitOp;
	TempOp.type = o_displ;
	TempOp.reg = BaseRegNum;
	TempOp.addr = (ea_t) offset;
	return dynamic_pointer_cast<STARS_op_t>(std::make_shared<STARS_IDA_op_t>(TempOp));
} // end of STARS_IDA_Instruction_t::MakeMemDisplacementOpnd()


